/*
 * Copyright 2019 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.kafka.secretregistry.storage.serialization;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;

import java.io.IOException;
import java.util.Map;

import io.confluent.kafka.secretregistry.storage.NoopKey;
import io.confluent.kafka.secretregistry.storage.SecretKey;
import io.confluent.kafka.secretregistry.storage.SecretRegistryKey;
import io.confluent.kafka.secretregistry.storage.SecretRegistryKeyType;
import io.confluent.kafka.secretregistry.storage.SecretRegistryValue;
import io.confluent.kafka.secretregistry.storage.SecretValue;
import io.confluent.kafka.secretregistry.storage.exceptions.SerializationException;

public class SecretRegistrySerializer
    implements Serializer<SecretRegistryKey, SecretRegistryValue> {

  public SecretRegistrySerializer() {

  }

  /**
   * @param key Typed key
   * @return bytes of the serialized key
   */
  @Override
  public byte[] serializeKey(SecretRegistryKey key) throws SerializationException {
    try {
      return new ObjectMapper().writeValueAsBytes(key);
    } catch (IOException e) {
      throw new SerializationException(
          "Error while serializing secret key" + key.toString(),
          e
      );
    }
  }

  /**
   * @param value Typed value
   * @return bytes of the serialized value
   */
  @Override
  public byte[] serializeValue(SecretRegistryValue value) throws SerializationException {
    try {
      return new ObjectMapper().writeValueAsBytes(value);
    } catch (IOException e) {
      throw new SerializationException(
          "Error while serializing value secret value " + value.toString(),
          e
      );
    }
  }

  @Override
  public SecretRegistryKey deserializeKey(byte[] key) throws SerializationException {
    SecretRegistryKey secretKey = null;
    SecretRegistryKeyType keyType = null;
    try {
      try {
        Map<Object, Object> keyObj = null;
        keyObj = new ObjectMapper().readValue(
            key,
            new TypeReference<Map<Object, Object>>() {
            }
        );
        keyType = SecretRegistryKeyType.forName((String) keyObj.get("keytype"));
        if (keyType == SecretRegistryKeyType.NOOP) {
          secretKey = new ObjectMapper().readValue(key, NoopKey.class);
        } else if (keyType == SecretRegistryKeyType.SECRET) {
          secretKey = new ObjectMapper().readValue(key, SecretKey.class);
          validateMagicByte((SecretKey) secretKey);
        }
      } catch (JsonProcessingException e) {

        String type = "unknown";
        if (keyType != null) {
          type = keyType.name();
        }

        throw new SerializationException("Failed to deserialize " + type + " key", e);
      }
    } catch (IOException e) {
      throw new SerializationException("Error while deserializing secret key", e);
    }
    return secretKey;
  }


  /**
   * @param key   Typed key corresponding to this value
   * @param value Bytes of the serialized value
   * @return Typed deserialized value.
   */
  @Override
  public SecretRegistryValue deserializeValue(SecretRegistryKey key, byte[] value)
      throws SerializationException {
    SecretRegistryValue secretRegistryValue = null;
    if (key.getKeyType().equals(SecretRegistryKeyType.SECRET)) {
      try {
        validateMagicByte((SecretKey) key);
        secretRegistryValue = new ObjectMapper().readValue(value, SecretValue.class);
      } catch (IOException e) {
        throw new SerializationException("Error while deserializing secret", e);
      }
    } else {
      throw new SerializationException("Unrecognized key type. Must be one of secret or config");
    }
    return secretRegistryValue;
  }

  @Override
  public void close() {

  }

  @Override
  public void configure(Map<String, ?> stringMap) {

  }

  private void validateMagicByte(SecretKey secretKey) throws SerializationException {
    if (secretKey.getMagicByte() != 0 && secretKey.getMagicByte() != 1) {
      throw new SerializationException("Can't deserialize secret for the magic byte "
          + secretKey.getMagicByte());
    }
  }
}
